#include "MathFuncs.hpp"


/*********************************************************************
 Weights for gaussian integration - from PHBPHD
********************************************************************/

/* ****************** Triangles: barycentric coordinates ******************* */
/* weights sum to 1. instead of 0.5, which is the area of the triangle
 * 0 <= x+y <= 1, x,y >= 0 */

/* degree 1, 1 point */
gaussTriangleWeights  gtrw1 = {
  1, 1,
  { 1./3. },
  { 1./3. },
  { 1.0	}
};


/* degree 2, 3 points, 
 * Stroud '71 p 307 */
gaussTriangleWeights gtrw2 = {
	2, 3,
        { 1./6.	, 1./6.	, 2./3.	},
        { 1./6.	, 2./3.	, 1./6.	},
        { 1./3.	, 1./3.	, 1./3.	}
};

/* degree 3, 4 points, 
 * Stroud '71 p 308 */
gaussTriangleWeights gtrw3 = {
	3, 4,
        {   1./3.,   0.2  ,   0.2  ,   0.6   },
        {   1./3.,   0.2  ,   0.6  ,   0.2   },
        { -9./16., 25./48., 25./48., 25./48. }
};

/* degree 3, 4 points, conical product Gauss rule 
gaussTriangleWeights gtrw3PG = {
	3, 4,

};
*/

/* degree 4, 6 points
 * Lyness, Jespersen, "Moderate Degree Symmertic Quadrature Rules for the 
 * Triangle", J. Inst. Maths. Applics (1975) 15, 19-32 */
#define w1	3.298552309659655e-1/3.
#define a1	8.168475729804585e-1
#define b1	9.157621350977073e-2
#define c1	b1
#define w2	6.701447690340345e-1/3.
#define a2	1.081030181680702e-1
#define b2	4.459484909159649e-1
#define c2	b2
gaussTriangleWeights gtrw4 = {
	4, 6, 
        { a1	, b1	, c1	, a2	, b2	, c2	},
        { b1	, c1	, a1	, b2	, c2	, a2	},
        { w1	, w1	, w1	, w2	, w2	, w2	}
};
#undef c2
#undef b2
#undef a2
#undef w2
#undef c1
#undef b1
#undef a1
#undef w1

/* degree 5, 7 points,
 * Stroud '71 p 314 */
#define r	0.1012865073234563
#define s	0.7974269853530873
#define t	1./3.
#define u	0.4701420641051151
#define v	0.05971587178976981
#define A	0.225
#define B	0.1259391805448271
#define C	0.1323941527885062
gaussTriangleWeights gtrw5 = {
	5, 7, 
        { t	, r	, r	, s	, u	, u	, v	},
        { t	, r	, s	, r	, u	, v	, u	},
        { A	, B	, B	, B	, C	, C	, C	}
};
#undef C
#undef B
#undef A
#undef v
#undef u
#undef t
#undef s
#undef r

/* degree 5, 9 points conical product Gauss rule 
gaussTriangleWeights gtrw5PG = {
	5, 9,

};
*/

/* degree 6: minimal number of nodes is 10, but there is at this time no rule
 * known with less than 12 nodes (see Cools & Rabanowitz). There is a rule of 
 * degree 7 which has only 12 nodes, so use that one instead. */

/* degree 7, 12 nodes,
 * Gaterman, "The Construction of Symmetric Cubature Formulas for the 
 * Square and the triangle", Computing, 40, 229-240 (1988) */
#define w1	0.2651702815743450e-1 * 2.
#define a1	0.6238226509439084e-1
#define b1	0.6751786707392436e-1
#define c1	0.8700998678316848
#define w2	0.4388140871444811e-1 * 2.
#define a2	0.5522545665692000e-1
#define b2	0.3215024938520156
#define c2	0.6232720494910644
#define w3	0.2877504278497528e-1 * 2.
#define a3	0.3432430294509488e-1
#define b3	0.6609491961867980
#define c3	0.3047265008681072
#define w4	0.6749318700980879e-1 * 2.
#define a4	0.5158423343536001
#define b4	0.2777161669764050
#define c4	0.2064414986699949
gaussTriangleWeights gtrw7 = {
	7, 12, 
        { a1	, b1	, c1	, a2	, b2	, c2	, 
	  a3	, b3	, c3	, a4	, b4	, c4	},
        { b1	, c1	, a1	, b2	, c2	, a2	,
	  b3	, c3	, a3	, b4	, c4	, a4	},
        { w1	, w1	, w1	, w2	, w2	, w2	,
	  w3	, w3	, w3	, w4	, w4	, w4	}
};
#undef c4
#undef b4
#undef a4
#undef w4
#undef c3
#undef b3
#undef a3
#undef w3
#undef c2
#undef b2
#undef a2
#undef w2
#undef c1
#undef b1
#undef a1
#undef w1

/* degree 7, 16 points conical product Gauss rule 
gaussTriangleWeights gtrw7PG = {
	7, 16, 

};
*/

/* degree 8, 16 points 
 * Lyness & Jespersen */
#define w0	1.443156076777862e-1	
#define a0	3.333333333333333e-1
#define b0	3.333333333333333e-1
#define	c0	b0
#define w1	2.852749028018549e-1/3.
#define a1	8.141482341455413e-2
#define b1	4.592925882927229e-1
#define c1	b1
#define w2	9.737549286959440e-2/3.
#define a2	8.989055433659379e-1
#define b2	5.054722831703103e-2
#define c2	b2
#define w3	3.096521116041552e-1/3.
#define a3	6.588613844964797e-1
#define b3	1.705693077517601e-1
#define c3	b3
#define w4	1.633818850466092e-1/6.
#define a4	8.394777409957211e-3
#define b4	7.284923929554041e-1
#define c4 	2.631128296346387e-1
gaussTriangleWeights gtrw8 = {
	8, 16, 
        { a0	, 
	  a1	, b1	, c1	,
	  a2	, b2	, c2	,
	  a3	, b3	, c3	,
	  a4	, a4	, b4	, b4	, c4	, c4	},
        { b0	, 
	  b1	, c1	, a1	,
	  b2	, c2	, a2	,
	  b3	, c3	, a3	,
	  b4	, c4	, c4	, a4	, a4	, b4	},
        { w0	,
	  w1	, w1	, w1	,
	  w2	, w2	, w2	,
	  w3	, w3	, w3	,
	  w4	, w4	, w4	, w4	, w4	, w4   }
};
#undef c4
#undef b4
#undef a4
#undef w4
#undef c3
#undef b3
#undef a3
#undef w3
#undef c2
#undef b2
#undef a2
#undef w2
#undef c1
#undef b1
#undef a1
#undef w1
#undef c0
#undef b0
#undef a0
#undef w0

/* degree 9, 19 points - there is as yet no rule over the triangle known 
 * which has only the minimal number of 17 nodes (see Cools & Rabinowitz). 
 * Lyness & Jespersen */
#define w0	9.713579628279610e-2   
#define a0	3.333333333333333e-1
#define b0	3.333333333333333e-1
#define c0	b0
#define w1	9.400410068141950e-2/3.
#define a1	2.063496160252593e-2
#define b1	4.896825191987370e-1
#define c1	b1
#define w2	2.334826230143263e-1/3.
#define a2	1.258208170141290e-1
#define b2	4.370895914929355e-1
#define c2	b2
#define w3	2.389432167816273e-1/3.
#define a3	6.235929287619356e-1
#define b3	1.882035356190322e-1
#define c3	b3
#define w4	7.673302697609430e-2/3.
#define a4	9.105409732110941e-1
#define b4	4.472951339445297e-2
#define c4	b4
#define w5	2.597012362637364e-1/6.
#define a5	3.683841205473626e-2
#define b5	7.411985987844980e-1	
#define c5	2.219629891607657e-1
gaussTriangleWeights gtrw9 = {
	9, 19, 
        { a0	, 
	  a1	, b1	, c1	,
	  a2	, b2	, c2	,
	  a3	, b3	, c3	,
	  a4	, b4	, c4	,
	  a5	, a5	, b5	, b5	, c5	, c5	},
        { b0	, 
	  b1	, c1	, a1	,
	  b2	, c2	, a2	,
	  b3	, c3	, a3	,
	  b4	, c4	, a4	,
	  b5	, c5	, c5	, a5	, a5	, b5	},
        { w0	,
	  w1	, w1	, w1	,
	  w2	, w2	, w2	,
	  w3	, w3	, w3	,
	  w4	, w4	, w4	,
	  w5	, w5	, w5	, w5	, w5	, w5   }
};
#undef c5
#undef b5
#undef a5
#undef w5
#undef c4
#undef b4
#undef a4
#undef w4
#undef c3
#undef b3
#undef a3
#undef w3
#undef c2
#undef b2
#undef a2
#undef w2
#undef c1
#undef b1
#undef a1
#undef w1
#undef c0
#undef b0
#undef a0
#undef w0

gaussTriangleWeights gtrw[10];
void initIntegration(){
  gtrw[0]=gtrw1;
  gtrw[1]=gtrw1;
  gtrw[2]=gtrw2;
  gtrw[3]=gtrw3;
  gtrw[4]=gtrw4;
  gtrw[5]=gtrw5;
  gtrw[6]=gtrw7;
  gtrw[7]=gtrw7;
  gtrw[8]=gtrw8;
  gtrw[9]=gtrw9;
}
