#include "MultipoleExpansionCoefficients.hpp"
#include "LocalExpansionCoefficients.hpp"

MultipoleExpansionCoefficients::MultipoleExpansionCoefficients(Vec3f r,Vec3f nd,float radiosity,Vec3f _center,int _p)
:ExpansionCoefficients(_center,_p) {
	r=r-center;
	Vec3f rs=sph(r);
	SourceMatrices sm(r,nd);
	fillYnmCache(rs[1],rs[2],p);
	for(int n=0;n<=p;n++){
		for(int j=0;j<=n/2;j++){
			for(int m=-n+2*j;m<=n-2*j;m++){
				csms[n][j][m]=radiosity*pow(rs[0],n)*conj(Ynm[n-2*j][m])*sm;
			}
		}
	}
}


MultipoleExpansionCoefficients MultipoleExpansionCoefficients::translateTo(Vec3f newCenter){
	MultipoleExpansionCoefficients m(newCenter,p);
	Vec3f ood=newCenter-center;
	Vec3f oods=sph(ood);
	TranslationMatrices tm(ood);
	Cplxf ans;
	fillYnmCache(oods[1],oods[2],p);
	for(int nd=0;nd<=p;nd++){
		for(int jd=0;jd<=nd/2;jd++){
			for(int md=-nd+2*jd;md<=nd-2*jd;md++){
				for(int k=0;k<=nd;k++){
					for(int s=0;s<=k/2;s++){
						for(int m1=-k+2*s;m1<=k-2*s;m1++){
							int jmin=max(jd-k+s,0);
							int jmax=min(jd-s,nd-k+s-jd);
							for(int j=jmin;j<=jmax;j++){
								if ((abs(m1)>(k-2*s))||(abs(md+m1)>(nd-k-2*j))) break;
								m.csms[nd][jd][md]+=(pow(oods[0],k)*Ynm[k-2*s][m1]*
									J(k-2*s,m1,nd-2*jd,s,-nd+k-4,md+m1,nd-k-2*j)*
									(e[nd-k][j]/e[nd][jd])*
									csms[nd-k][j][m1+md]);
								
							}
						}
					}
				}
				m.csms[nd][jd][md]=tm*m.csms[nd][jd][md];
			}
		}
	}
	return m;
}

float MultipoleExpansionCoefficients::evaluateAtPoint(Vec3f r,Vec3f nd,float rho){
	r=r-center;
	ReceiverMatrices rm(r,nd);
	Vec3f rs=sph(r);
	Cplxf ans(0,0);
	fillYnmCache(rs[1],rs[2],p);
	for(int n=0;n<=p;n++){
		for(int j=0;j<=n/2;j++){
			for(int m=-n+2*j;m<=n-2*j;m++){
				CplxReceiverMatrices a=pow(rs[0],-n-4)*Ynm[n-2*j][m]*rm;
				ans+=e[n][j]*a*csms[n][j][m];
			}
		}
	}
	return rho*real(ans);
}


LocalExpansionCoefficients MultipoleExpansionCoefficients::getLocalCoefficientsFor(Vec3f newCenter){
	LocalExpansionCoefficients l(newCenter,p);
	Vec3f ox0=newCenter-center;
	Vec3f ox0s=sph(ox0);
	TranslationMatrices tm(ox0);
	fillYnmCache(ox0s[1],ox0s[2],2*p);
	for(int k=0;k<=p;k++){
		for(int s=0;s<=k/2;s++){
			for(int m1=-k+2*s;m1<=k-2*s;m1++){
				for(int n=0;n<=p;n++){
					for(int j=0;j<=n/2;j++){
						for(int m=-n+2*j;m<=n-2*j;m++){
							int lambdamin=max(abs(n-2*j-k+2*s),abs(m-m1));
							int lmax=n-2*j+k-2*s;
							int lmin=((n-2*j-k+2*s-lambdamin)%2==0?(lambdamin):(lambdamin+1));
							for(int l2=lmin;l2<=lmax;l2+=2){
								l.csms[k][s][m1]+=tm*(J(k-2*s,m1,l2,s,-n-4,m,n-2*j)*e[n][j]*pow(ox0s[0],-n-4-k)*Ynm[l2][m-m1]*csms[n][j][m]);
							}
						}
					}
				}
			}
		}
	}
	return l;
	
}



void MultipoleExpansionCoefficients::addTranslatedMultipole(MultipoleExpansionCoefficients* mecs){
	Vec3f ood=center-mecs->center;
	Vec3f oods=sph(ood);
	TranslationMatrices tm(ood);
	Cplxf ans;
	fillYnmCache(oods[1],oods[2],p);
	for(int nd=0;nd<=p;nd++){
		for(int jd=0;jd<=nd/2;jd++){
			for(int md=-nd+2*jd;md<=nd-2*jd;md++){
				CplxSourceMatrices cndjdmd;
				for(int k=0;k<=nd;k++){
					for(int s=0;s<=k/2;s++){
						for(int m1=-k+2*s;m1<=k-2*s;m1++){
							int jmin=max(jd-k+s,0);
							int jmax=min(jd-s,nd-k+s-jd);
							for(int j=jmin;j<=jmax;j++){
								if ((abs(m1)>(k-2*s))||(abs(md+m1)>(nd-k-2*j))) break;
								  cndjdmd+=(pow(oods[0],k)*Ynm[k-2*s][m1]*
									J(k-2*s,m1,nd-2*jd,s,-nd+k-4,md+m1,nd-k-2*j)*
									(e[nd-k][j]/e[nd][jd])*
									mecs->csms[nd-k][j][m1+md]);
								
							}
						}
					}
				}
				csms[nd][jd][md]+=tm*cndjdmd;
			}
		}
	}
}

void MultipoleExpansionCoefficients::addPatchPointMultipole(PatchPoint* pp,int channel){
	Vec3f r=pp->coords-center;
	Vec3f rs=sph(r);
	SourceMatrices sm(r,pp->normal);
	fillYnmCache(rs[1],rs[2],p);
	for(int n=0;n<=p;n++){
		for(int j=0;j<=n/2;j++){
			for(int m=-n+2*j;m<=n-2*j;m++){
				csms[n][j][m]+=pp->weight*pp->radVal[channel]*pow(rs[0],n)*conj(Ynm[n-2*j][m])*sm;
			}
		}
	}

}
